import smtplib
import config

def send_mail(to, subject, msg):
    program = '/usr/sbin/sendmail'
    email = os.popen("%s -t" % program, "w")
    email.write("From: %s\n" % config.email_user)
    email.write("Reply-to: %s\n" % config.email_user)
    email.write("To: %s\n" % to)
    email.write("Subject: %s\n" % subject)
    email.write("\n")
    email.write("%s\n" % msg)
    status = email.close()


def send_notification(notify_list, notification_type, notifier=send_mail):
    SUBJECT = 'OSM Hall Monitor Notification |'
    messages = {}
    subjects = {}
    tos = {}

    for entry in notify_list:
        if notification_type == 'user':
            if entry[6]:
                MSG = """
Dear %s,
OSM Hall Monitor has detected an event for your consideration.

  **User alert**
       Time of event: %s
       Username: http://www.osm.org/user/%s
       Changeset: http://www.osm.org/changeset/%s
       Additions: %s
       Modifications: %s
       Deletions: %s
       Reason user is watched: %s

Problem? Feedback? Reply to this message.

Best,

OSM Hall Monitor
""" % (entry[4], entry[0][0], entry[0][2], entry[0][1], entry[0][3], entry[0][4], entry[0][5], entry[3])

                TO = entry[6]
                NEWSUBJECT = '%s User %s ' % (SUBJECT, entry[0][2])

        elif notification_type == 'object':
            if entry[6]:
                if 'n' == entry[0][4][0]:
                    pre = 'node'
                elif 'w' == entry[0][4][0]:
                    pre = 'way'
                elif 'r' == entry[0][4][0]:
                    pre = 'relation'
                MSG = """
Dear %s,
OSM Hall Monitor has detected an event for your consideration.

  **Object alert**
       Time of event: %s
       Object: http://www.osm.org/%s/%s
       Changeset: http://www.osm.org/changeset/%s
       Action: %s
       User performing: http://www.osm.org/user/%s
       Reason object is watched: %s

Problem? Feedback? Reply to this message.

Best,

OSM Hall Monitor
""" % (entry[4], entry[0][0], pre, entry[0][4][1:], entry[0][1], entry[0][3], entry[0][2], entry[3])

                TO = entry[6]
                NEWSUBJECT = '%s Object %s ' % (SUBJECT, entry[0][4])

        else:
            continue

        addr = TO
        if addr not in tos:
            tos[addr] = TO
            subjects[addr] = NEWSUBJECT
            messages[addr] = MSG
        else:
            subjects[addr] = '%s: multiple events' % (SUBJECT)
            messages[addr] += '\r\n---NEXT ALERT---\r\n'+MSG

    for email in messages:
        try:
            notifier(tos[email], subjects[email], messages[email])
        except Exception,e:
            print 'Issue sending notification: ', str(e)
            pass



import connect
import inserts
import send_notification
import queries
import fnmatch


def suspicious_filter(changesets):
    """Set of rudimentary filters towards detecting possibly bad changesets.    

    1: Large amount of additions
    2: Large amount of modifications
    3: Large amount of deletions
    4: Large amount of combined actions
    5: High proportion of deletions
    6: High proportion of modifications
    """
    whitelist = queries.query_white_list()

    conn = connect.connect()
    cur = conn.cursor()

    for changesetid, changeset in changesets.iteritems():
        if changeset['username'] in whitelist:
            continue
        if changeset['create'] > 1500:
            info = (changeset['timestamp'], changesetid,
                    changeset['username'].encode('utf8'),
                    1, changeset['create'])
            cur.execute("""INSERT INTO history_filters
                            (timestamp,changeset,username,flag,quantity)
                            VALUES (%s, %s, %s, %s, %s);""", info)
        if changeset['modify'] > 1500:
            info = (changeset['timestamp'], changesetid,
                    changeset['username'].encode('utf8'),
                    2, changeset['modify'])
            cur.execute("""INSERT INTO history_filters
                            (timestamp,changeset,username,flag,quantity)
                            VALUES (%s, %s, %s, %s, %s);""", info)
        if changeset['delete'] > 1500:
            info = (changeset['timestamp'], changesetid,
                    changeset['username'].encode('utf8'),
                    3, changeset['delete'])
            cur.execute("""INSERT INTO history_filters
                            (timestamp,changeset,username,flag,quantity)
                            VALUES (%s, %s, %s, %s, %s);""", info)
        if changeset['create'] + changeset['modify'] + changeset['delete'] > 1500:
            info = (changeset['timestamp'], changesetid,
                    changeset['username'].encode('utf8'),
                    4, changeset['create'] + changeset['modify'] + changeset['delete'])
            cur.execute("""INSERT INTO history_filters
                            (timestamp,changeset,username,flag,quantity)
                            VALUES (%s, %s, %s, %s, %s);""", info)
        if changeset['delete'] > 0 and float(changeset['create']+changeset['modify'])/float(changeset['delete']) < 0.001:
            info = (changeset['timestamp'], changesetid,
                    changeset['username'].encode('utf8'),
                    5, float(changeset['create']+changeset['modify'])/float(changeset['delete']))
            cur.execute("""INSERT INTO history_filters
                            (timestamp,changeset,username,flag,quantity)
                            VALUES (%s, %s, %s, %s, %s);""", info)
        if changeset['modify'] > 0 and float(changeset['create']+changeset['delete'])/float(changeset['modify']) < 0.001:
            info = (changeset['timestamp'], changesetid,
                    changeset['username'].encode('utf8'),
                    6, float(changeset['create']+changeset['delete'])/float(changeset['modify']))
            cur.execute("""INSERT INTO history_filters
                            (timestamp,changeset,username,flag,quantity)
                            VALUES (%s, %s, %s, %s, %s);""", info)

    conn.commit()


def user_filter(changesets, notification=False, notifier=send_notification.send_mail):
    notify_list = []

    watched_users = queries.query_user_list()

    conn = connect.connect()
    cur = conn.cursor()

    if watched_users:
        for changesetid, changeset in changesets.iteritems():
            for user in watched_users:
                if fnmatch.fnmatch(changeset['username'].encode('utf-8'), user['username']):
                    inserts.insert_user_event(changeset, user['id'])

                    notify_list.append([(changeset['timestamp'], changesetid, changeset['username'].encode('utf8'),
                                         changeset['create'], changeset['modify'], changeset['delete'])] + user)
    if notify_list and notification:
        send_notification.send_notification(notify_list, 'user', notifier=notifier)


def user_object_filter(objects, notification=False, notifier=send_notification.send_mail):
    notify_list = []

    watched_users = queries.query_user_object_list()

    conn = connect.connect()
    cur = conn.cursor()

    if watched_users:
        for user in watched_users:
            for item_id, item in objects.iteritems():
                if fnmatch.fnmatch(item['username'].encode('utf-8'), user['username']):
                    if item['create'] == 1:
                        action = 'create'
                    elif item['modify'] == 1:
                        action = 'modify'
                    elif item['delete'] == 1:
                        action = 'delete'
                    for item_key in item['tags']:
                        info = (item['timestamp'], item['changeset'],
                        item['username'].encode('utf8'), action,
                        item_key, item['tags'][item_key])
                        cur.execute("""INSERT INTO history_users_objects
                                    (timestamp,changeset,username,action,key,value)
                                    VALUES (%s, %s, %s, %s, %s, %s);""", info)
        conn.commit()


def object_filter(objects, notification=False, notifier=send_notification.send_mail):
    notify_list = []

    watched_objects = queries.query_object_list()

    conn = connect.connect()
    cur = conn.cursor()

    if watched_objects:
        for obj in watched_objects:
            for item_id, item in objects.iteritems():
                if item_id == obj['element']:
					if item['create'] == 1:
						item['action'] = 1
					elif item['modify'] == 1:
						item['action'] = 2
					elif item['delete'] == 1:
						item['action'] = 4
                                        inserts.insert_object_event(item, obj['id'])

					notify_list.append([(item['timestamp'], item['changeset'], item['username'].encode('utf8'),
                                                             item['action'], item_id)] + obj)
    if notify_list and notification:
        send_notification.send_notification(notify_list, 'object', notifier=notifier)


def key_filter(objects, notification=False, notifier=send_notification.send_mail):
    notify_list = []

    watched_keys = queries.query_key_list()

    conn = connect.connect()
    cur = conn.cursor()

    if watched_keys:
        for key in watched_keys:
            for item_id, item in objects.iteritems():
                for item_key in item['tags']:
                    if fnmatch.fnmatch(item_key,key['key']) and fnmatch.fnmatch(item['tags'][item_key],key['value']):
                        if item['create'] == 1:
                            item['action'] = 1
                        elif item['modify'] == 1:
                            item['action'] = 2
                        elif item['delete'] == 1:
                            item['action'] = 4
                        inserts.insert_key_event(item, item_key, key['id'])

                        notify_list.append([(item['timestamp'], item['changeset'], item['username'].encode('utf8'),
                                             item['action'], item_key, item['tags'][item_key])])
    if notify_list and notification:
        send_notification.send_notification(notify_list, 'key', notifier=notifier)



"""
queries.py

Retrieves lists of entries on tracking lists and do-not-track lists.

"""
import connect


def query_white_list():
    """
    Retrieve list of users on whitelist.

    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("SELECT username FROM whitelisted_users")
    white_list = cur.fetchall()
    whitelist = [name[0] for name in white_list]

    return white_list


def query_user_list():
    """
    Retrieve list of users on user watch list.

    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("SELECT * FROM watched_users")
    watched_users = cur.fetchall()

    return watched_users


def query_user_object_list():
    """
    Retrieve list of users on object composite watch list.

    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("SELECT * FROM watched_users_objects")
    watched_users = cur.fetchall()

    return watched_users


def query_object_list():
    """
    Retrieve list of objects on object watch list.

    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("SELECT * FROM watched_objects")
    watched_objects = cur.fetchall()

    return watched_objects


def query_key_list():
    """
    Retrieve list of key/value pairs on key/value pair list.

    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("SELECT * FROM watched_keys")
    watched_keys = cur.fetchall()

    return watched_keys



"""
fetch.py

Contains functions to fetch diff file information (state files) from
  OpenStreetMap (OSM) planet server.

"""
import config
from connect import connect
import requests


def fetch_last_read():
    """
    Accesses the file_list table to retrieve the most recently read
      diff file as well as information about it. If the file listing
      does not contain any information, it returns None.

    """
    conn = connect()
    cur = conn.cursor()

    cur = conn.cursor()
    cur.execute("SELECT * FROM file_list;")

    try:
        sequence = {}
        _, sequence['sequencenumber'], sequence['timestamp'], \
            sequence['timetype'], sequence['read_flag'] = cur.fetchone()

        return sequence
    except:
        return None


def fetch_next(current_sequence='', time_type='hour', reset=False):
    """
    Given a current sequence, fetches information about the next
      diff file from its state file. If no sequence has been read
      before, can fetch information for the most recent one. Once
      fetched, file_list table is updated with this information.

    Inputs
    ------
    current_sequence : float or int
        Up to nine digit number corresponding to the last read
          sequence by the system. Inputs less than 9 digits are
          left-padded with zeros.

    time_type : {'minute', 'hour', 'day'}
        Time resolution of diff files desired.

    reset : bool
        Flag to retrieve latest diff sequence state from OSM and
          clear old file listing in database if it exists.

    """
    if reset is True:
        state_url = ("http://planet.openstreetmap.org/replication/"
                     "%s/state.txt") % (time_type)

    else:
        next_sequence = int(current_sequence) + 1
        sequence_string = str(next_sequence).zfill(9)
        state_url = ("http://planet.openstreetmap.org/replication/"
                     "%s/%s/%s/%s.state.txt") % (time_type,
                         sequence_string[0:3], sequence_string[3:6],
                         sequence_string[6:9])

    if time_type == 'minute':
        end = 5
    else:
        end = 3

    response = requests.get(state_url, headers=config.http_headers)

    if response.status_code == 404:
        raise Exception('Unable to reach OpenStreetMap server.')

    state = {}
    vals = response.text.encode('utf-8').split('\n')

    for val in vals[1:end]:
        (k, v) = val.split('=')
        state[k.lower()] = v.strip().replace("\\:", ":")

    info = (state['sequencenumber'], state['timestamp'], time_type, False)

    conn = connect()
    cur = conn.cursor()

    if reset is True:
        cur.execute("DELETE FROM file_list;")
        cur.execute("""INSERT INTO file_list
                       (sequence, timestamp, timetype, read)
                       VALUES (%s, %s, %s, %s);""", info)
    else:
        cur.execute("""UPDATE file_list SET
                       (sequence, timestamp, timetype, read)
                       = (%s, %s, %s, %s);""", info)

    conn.commit()



import connect


error_message = 'Action is not defined. Please use create, truncate, or drop.'
def file_list(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE file_list (
              id SERIAL NOT NULL PRIMARY KEY,
              sequence TEXT,
              timestamp TEXT,
              timetype TEXT,
              read BOOLEAN
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
              TRUNCATE TABLE file_list;
              """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
              DROP TABLE file_list;
              """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def history_users(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE history_users (
              id SERIAL NOT NULL PRIMARY KEY,
              wid INTEGER NOT NULL,
              userid BIGINT NOT NULL,
              changeset BIGINT NOT NULL,
              timestamp TEXT NOT NULL,
              created BIGINT,
              modified BIGINT,
              deleted BIGINT
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE history_users;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE history_users;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def history_all_users(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE history_all_users (
              id SERIAL NOT NULL PRIMARY KEY,
              username TEXT NOT NULL,
              changeset TEXT NOT NULL,
              timestamp TEXT NOT NULL,
              created TEXT,
              modified TEXT,
              deleted TEXT
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE history_all_users;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE history_all_users;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def history_all_changesets(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE history_all_changesets (
              id SERIAL NOT NULL PRIMARY KEY,
              changeset TEXT NOT NULL,
              username TEXT NOT NULL,
              timestamp TEXT NOT NULL,
              created TEXT,
              modified TEXT,
              deleted TEXT
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE history_all_changesets;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE history_all_changesets;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def history_users_objects(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE history_users_objects (
              id SERIAL NOT NULL PRIMARY KEY,
              key TEXT NOT NULL,
              value TEXT NOT NULL,
              username TEXT NOT NULL,
              changeset BIGINT NOT NULL,
              timestamp TEXT NOT NULL,
              action TEXT
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE history_users_objects;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE history_users_objects;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def history_objects(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE history_objects (
              id SERIAL NOT NULL PRIMARY KEY,
              wid INTEGER NOT NULL,
              userid BIGINT NOT NULL,
              username TEXT NOT NULL,
              changeset BIGINT NOT NULL,
              timestamp TEXT NOT NULL,
              action SMALLINT NOT NULL
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE history_objects;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE history_objects;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def history_keys(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE history_keys (
              id SERIAL NOT NULL PRIMARY KEY,
              wid INTEGER NOT NULL,
              userid BIGINT NOT NULL,
              key TEXT NOT NULL,
              value TEXT NOT NULL,
              element TEXT NOT NULL,
              username TEXT NOT NULL,
              changeset BIGINT NOT NULL,
              timestamp TEXT NOT NULL,
              action SMALLINT NOT NULL
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE history_keys;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE history_keys;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def history_filters(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE history_filters (
              id SERIAL NOT NULL PRIMARY KEY,
              flag INT NOT NULL,
              username TEXT NOT NULL,
              changeset BIGINT NOT NULL,
              timestamp TEXT NOT NULL,
              quantity TEXT NOT NULL
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE history_filters;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE history_filters;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def watched_users(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE watched_users (
              id SERIAL NOT NULL PRIMARY KEY,
              userid BIGINT,
              username TEXT NOT NULL,
              reason TEXT,
              author TEXT,
              authorid BIGINT,
              email TEXT
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE watched_users;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE watched_users;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def watched_users_objects(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE watched_users_objects (
              id SERIAL NOT NULL PRIMARY KEY,
              username TEXT NOT NULL,
              reason TEXT,
              author TEXT,
              authorid BIGINT,
              email TEXT
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE watched_users_objects;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE watched_users_objects;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def watched_objects(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE watched_objects (
              id SERIAL NOT NULL PRIMARY KEY,
              element TEXT NOT NULL,
              reason TEXT,
              author TEXT,
              authorid BIGINT,
              email TEXT
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE watched_objects;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE watched_objects;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def watched_keys(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE watched_keys (
              id SERIAL NOT NULL PRIMARY KEY,
              key TEXT NOT NULL,
              value TEXT NOT NULL,
              reason TEXT,
              author TEXT,
              authorid BIGINT,
              email TEXT
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE watched_keys;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE watched_keys;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def whitelisted_users(action):
    conn = connect.connect()
    cur = conn.cursor()
    if action in ['create', 'c']:
        cur.execute("""
            CREATE TABLE whitelisted_users (
              id SERIAL NOT NULL PRIMARY KEY,
              username TEXT NOT NULL,
              reason TEXT,
              author TEXT,
              authorid BIGINT
            );
            """)
    elif action in ['truncate', 't']:
        cur.execute("""
            TRUNCATE TABLE whitelisted_users;
            """)
    elif action in ['drop', 'delete', 'd']:
        cur.execute("""
            DROP TABLE whitelisted_users;
            """)
    else:
        raise NotImplementedError(error_message)
    conn.commit()


def all_tables(action):
    if action in ['create', 'c']:
        message = "You are about to create all tables for OSM Hall Monitor. Are you sure? (Y or N): "
    elif action in ['truncate', 't']:
        message = "You are about to truncate all tables but the file list for OSM Hall Monitor. Are you sure? (Y or N): "
        second_message = "Would you like to truncate the file list table as well? (Y or N): "
    elif action in ['drop', 'delete', 'd']:
        message = "You are about to drop all tables for OSM Hall Monitor. Are you sure? (Y or N): "
    else:
        raise NotImplementedError(error_message)

    response = str(raw_input(message))

    if response.lower() not in ['y', 'yes']:
        print 'Execution halted.'
        return

    if action in ['truncate', 't']:
        response = str(raw_input(second_message))
        if response.lower() not in ['y', 'yes']:
            pass
        else:
            file_list(action)
    else:
        file_list(action)

    history_users(action)
    history_users_objects(action)
    history_all_users(action)
    history_all_changesets(action)
    history_objects(action)
    history_keys(action)
    history_filters(action)
    watched_users(action)
    watched_users_objects(action)
    watched_objects(action)
    watched_keys(action)
    whitelisted_users(action)

    print 'Done.'



"""
manage.py

Functions that add or remove entries on tracking lists.

"""
import connect


def add_watched_user(username, reason, author, authorid, email):
    """
    Add user to watched user list for tracking.

    Inputs
    ------
    username : str
        Username to track
    reason : str
        Reason to track user
    author : str
        User adding tracking entry
    authorid : int
        Userid of user adding entry
    email : str, option
        Email address for notification of events
    
    """
    conn = connect.connect()
    cur = conn.cursor()

    info = (username, reason, author, authorid, email)

    cur.execute("""INSERT INTO watched_users
                   (username, reason, author, authorid, email)
                   VALUES (%s, %s, %s, %s, %s);""", info)

    conn.commit()


def remove_watched_user(username):
    """
    Remove user from tracking list.

    Inputs
    ------
    username : str
        Username to remove from database
    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("""DELETE FROM watched_users WHERE
                   username = %s;""", username)

    conn.commit()


def add_watched_user_object(username, reason, author, authorid, email):
    """
    Add user to watched user list with object composites for tracking.

    Inputs
    ------
     username : str
        Username to track
    reason : str
        Reason to track user
    author : str
        User adding tracking entry
    authorid : int
        Userid of user adding entry
    email : str, option
        Email address for notification of events
    
    """
    conn = connect.connect()
    cur = conn.cursor()

    info = (username, reason, author, authorid, email)

    cur.execute("""INSERT INTO watched_users_objects
                   (username, reason, author, authorid, email)
                   VALUES (%s, %s, %s, %s, %s);""", info)


def remove_watched_user_object(username):
    """
    Remove user from object composite user tracking list.
    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("""DELETE FROM watched_users_objects WHERE
                   username = %s;""", username)


def add_watched_object(element, reason, author, authorid, email):
    """
    Add object to watched object list.

    Inputs
    ------
    element : str
        Object to track, with type specified as single letter
          prepended to object id (e.g. node 322 is 'n322')
    reason : str
        Reason to track user
    author : str
        User adding tracking entry
    authorid : int
        Userid of user adding entry
    email : str, option
        Email address for notification of events
 
    """
    conn = connect.connect()
    cur = conn.cursor()

    info = (element, reason, author, authorid, email)

    cur.execute("""INSERT INTO watched_objects
                   (element, reason, author, authorid, email)
                   VALUES (%s, %s, %s, %s, %s);""", info)


    conn.commit()


def remove_watched_object(element):
    """
    Remove object from object tracking list.
    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("""DELETE FROM watched_objects WHERE
                   element = %s;""", element)

    conn.commit()


def add_watched_key(key, value, reason, author, authorid, email):
    """
    Add key/value combination to key/value tracking list.

    Inputs
    ------
    key : str
        Key to track; can be wildcard
    value : str
        Key value to track; can be wildcard
    reason : str
        Reason to track user
    author : str
        User adding tracking entry
    authorid : int
        Userid of user adding entry
    email : str, option
        Email address for notification of events
 
    """
    conn = connect.connect()
    cur = conn.cursor()

    info = (key, value, reason, author, authorid, email)

    cur.execute("""INSERT INTO watched_keys
                   (key, value, reason, author, authorid, email)
                   VALUES (%s, %s, %s, %s, %s, %s);""", info)

    conn.commit()


def remove_watched_key(key, value):
    """
    Remove key/value combination from key/value tracking list.

    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("""DELETE FROM watched_keys WHERE
                   key = %s AND value = %s;""", (key, value))

    conn.commit()


def add_whitelisted_user(username, reason, author, authorid):
    """
    Add whitelisted user that is not picked up in tracking.

    Inputs
    ------
    username : str
        Username to track
    reason : str
        Reason to track user
    author : str
        User adding tracking entry
    authorid : int
        Userid of user adding entry
 
    """
    conn = connect.connect()
    cur = conn.cursor()

    info = (username, reason, author, authorid)

    cur.execute("""INSERT INTO whitelisted_users
                   (username, reason, author, authorid)
                   VALUES (%s, %s, %s, %s);""", info)

    conn.commit()


def remove_whitelisted_user(username):
    """
    Remove whitelisted user from untracked list.

    """
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("""DELETE FROM whitelisted_users WHERE
                   username = %s;""", username)

    conn.commit()



"""
connect.py

Provides a connection to the database.

"""
import config
import psycopg2
from psycopg2.extras import DictCursor
import urlparse


def connect():
    """
    Creates a connection object for the database. This function uses
      the variable osmhm.config.database_url as the database address.
      Only PostgreSQL support is available at this time.

    """
    urlparse.uses_netloc.append('postgres')
    db_url = urlparse.urlparse(config.database_url)

    try:
        connection = psycopg2.connect(
            database=db_url.path[1:],
            user=db_url.username,
            password=db_url.password,
            host=db_url.hostname,
            port=db_url.port,
            cursor_factory=DictCursor,
            )
    except:
        raise Exception("""Error connecting to database! Please check
                           the database configuration variable.""")

    return connection



"""
config.py

Imports the local environment configuration parameters used in
  database access, event notification, and HTTP headers.

"""
import os
import warnings

http_headers = {'User-Agent': 'OSM Hall Monitor v0.3'}

try:
    database_url = os.environ['DATABASE_URL']
except:
    msg = 'No database_url detected; please set it with config.'
    warnings.warn(msg, UserWarning)

try:
    email_user = os.environ['EMAIL_USER']
except:
    msg = 'No email_user detected; please set it with config.'
    warnings.warn(msg, UserWarning)



import connect


def insert_file_read():
    conn = connect.connect()
    cur = conn.cursor()

    cur.execute("UPDATE file_list SET read = '%s';", (True,))
    conn.commit()


def insert_user_event(changeset, wid):
    conn = connect.connect()
    cur = conn.cursor()

    info = (wid, changeset['id'], changeset['timestamp'], changeset['uid'],
            changeset['create'], changeset['modify'], changeset['delete'])
    cur.execute("""INSERT INTO history_users
                    (wid, changeset, timestamp, userid, created, modified, \
                    deleted) VALUES (%s, %s, %s, %s, %s, %s, %s);""", info)
    conn.commit()


def insert_object_event(object, wid):
    conn = connect.connect()
    cur = conn.cursor()

    info = (wid, object['timestamp'],
            object['username'].encode('utf8'), object['uid'],
            object['action'], object['changeset'])
    cur.execute("""INSERT INTO history_objects
                    (wid, timestamp, username, userid, action, changeset)
                    VALUES (%s, %s, %s, %s, %s, %s);""", info)
    conn.commit()


def insert_key_event(object, key, wid):
    conn = connect.connect()
    cur = conn.cursor()

    info = (wid, object['id'], object['timestamp'],
            object['username'].encode('utf8'), object['uid'],
            object['action'], key, object['tags'][key],
            object['changeset'])
    cur.execute("""INSERT INTO history_keys
                    (wid, element, timestamp, username, userid, action, key, 
                    value, changeset) VALUES
                    (%s, %s, %s, %s, %s, %s, %s, %s, %s);""", info)
    conn.commit()


def insert_all_users(users):
    conn = connect.connect()
    cur = conn.cursor()

    for username, user in users.iteritems():
        info = (username, user["changesets"], user["timestamps"],
                    user["action"]["create"], user["action"]["modify"],
                    user["action"]["delete"])
        cur.execute("""INSERT INTO history_all_users
                        (username, changeset, timestamp, created, modified, deleted)
                        VALUES (%s, %s, %s, %s, %s, %s);""", info)

    conn.commit()


def insert_all_changesets(changesets):
    conn = connect.connect()
    cur = conn.cursor()

    for changesetid, changeset in changesets.iteritems():
        info = (changesetid, changeset["username"], changeset["timestamp"],
                    changeset["create"], changeset["modify"],
                    changeset["delete"])
        cur.execute("""INSERT INTO history_all_changesets
                       (changeset, username, timestamp, created, modified, deleted)
                       VALUES (%s, %s, %s, %s, %s, %s);""", info)

    conn.commit()



from setuptools import setup

setup(name='osm_hall_monitor',
      version='0.52.1',
      description='Passive changeset monitoring for OpenStreetMap.',
      url='http://github.com/ethan-nelson/osm_hall_monitor',
      author='Ethan Nelson',
      license='MIT',
      author_email='git@ethan-nelson.com',
      install_requires = ['psycopg2','osm_diff_tool'],
      packages=['osmhm'])
      



import fetch
import filters
import inserts
import tables
import config
import send_notification


def run(time_type='hour', history=False, suspicious=False, monitor=True,
        notification=False, notifier=send_notification.send_mail):
    """
    """
    import osmhm
    import osmdt
    import datetime
    import time

    while True:

        sequence = osmhm.fetch.fetch_last_read()

        if not sequence:
            osmhm.fetch.fetch_next(time_type=time_type, reset=True)
            sequence = osmhm.fetch.fetch_last_read()

        if sequence['read_flag'] is False:
            print "Processing sequence %s." % (sequence['sequencenumber'])

            count = 0
            while True:
                try:
                    count += 1
                    data_stream = osmdt.fetch(sequence['sequencenumber'], time=time_type)
                    break
                except:
                    if count == 5:
                        msg = 'Current state file not retrievable after five times.'
                        raise Exception(msg)
                    print "File not reachable; waiting 60 more seconds..."
                    time.sleep(60)

            data_object = osmdt.process(data_stream)
            del data_stream

            changesets = osmdt.extract_changesets(data_object)
            objects = osmdt.extract_objects(data_object)
            users = osmdt.extract_users(data_object)
            del data_object

            if history:
                osmhm.inserts.insert_all_changesets(changesets)

            if suspicious:
                osmhm.filters.suspicious_filter(changesets)

            if monitor:
                osmhm.filters.object_filter(objects, notification=notification, notifier=notifier)
                osmhm.filters.user_filter(changesets, notification=notification, notifier=notifier)
                #osmhm.filters.user_object_filter(objects, notification=notification, notifier=notifier)  # not implemented yet
                osmhm.filters.key_filter(objects, notification=notification, notifier=notifier)

            del changesets, objects, users

            osmhm.inserts.insert_file_read()
            print "Finished processing %s." % (sequence['sequencenumber'])

        if sequence['timetype'] == 'minute':
            delta_time = 1
            extra_time = 10
        elif sequence['timetype'] == 'hour':
            delta_time = 60
            extra_time = 120
        elif sequence['timetype'] == 'day':
            delta_time = 1440
            extra_time = 300

        next_time = datetime.datetime.strptime(sequence['timestamp'],
                      "%Y-%m-%dT%H:%M:%SZ") + datetime.timedelta(minutes=delta_time)

        if datetime.datetime.utcnow() < next_time:
            sleep_time = (next_time - datetime.datetime.utcnow()).seconds + delta_time
            print "Waiting %2.1f seconds for the next file." % (sleep_time)
        else:
            sleep_time = 1

        time.sleep(sleep_time)

        count = 0
        while True:
            try:
                count += 1
                osmhm.fetch.fetch_next(sequence['sequencenumber'], time_type=time_type)
                break
            except:
                if count == 5:
                    msg = 'New state file not retrievable after five times.'
                    raise Exception(msg)
                print "Waiting %2.1f more seconds..." % (extra_time)
                time.sleep(extra_time)
